#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<structure.h>
#include "lt.h"

/****************************************************************/
/*		nom:		cosmoratio			*/
/*		auteur:		Jean-Paul Kneib			*/
/*		date:		10/02/92			*/
/*		place:		Toulouse			*/
/****************************************************************/

static double gg(double z);
static double sk(double x,double k);
static double chi1(double z);
static double chi2(double z1,double z2);

/**************************************************************/

double distcosmo1(double z)
/* Return the angular distance DA(observator(z=0),object(z)) (no unit)
 * Multiply discosmo1(z) by c/H0 to get the true value in Mpc.
 * distcosmo1(z) * c/H0 = DA(0,z) = 1/(1+z) * Sk( integral( 0,z,c*dz/H(z) ) )
 * 
 * Global variables used : 
 * - C
 */
{
	extern struct g_cosmo C;
	double g;
	
	if (C.omegaX==0.)
	{
		g=gg(z);
		// Reformulation of the Mattig relation of OL = OK = 0 (De Sitter)
		return(2.*((1.-C.omegaM-g)*(1.-g))/C.omegaM/C.omegaM/(1.+z)/(1.+z));
	}
	else
	{
		if (C.kcourb!=0.)
			return(sk(chi1(z)*sqrt(fabs(C.kcourb)),C.kcourb)
				/(1+z)/sqrt(fabs(C.kcourb)));
		else
			return(chi1(z)/(1+z));
	}
}

/**************************************************************/
/* Return the angular distance DA(object(z1),object(z2)) divided by c/H0
 * distcosmo2(z1,z2) * c/H0 = DA(z1,z2) = 1/(1+z2) * Sk( integral( z1,z2,c*dz/H(z) ) )
 * 
 * Global variables used :
 * - C
 * - in gg() : C
 * - in chi2() : C
 */
double distcosmo2(double z1,double z2)
{
	extern struct	g_cosmo	C;
	double  g1,g2;
	
	if (C.omegaX==0.)
	{
		g1=gg(z1);
		g2=gg(z2);
		// Mattig relation for a De Sitter Universe
		return(2.*((1.-C.omegaM-g1*g2)*(g1-g2))
				/C.omegaM/C.omegaM/(1.+z1)/(1.+z2)/(1.+z2));
	}
	else
	{
		if( C.kcourb!=0. )
			return(sk(chi2(z1,z2)*sqrt(fabs(C.kcourb)),C.kcourb)
					/(1+z2)/sqrt(fabs(C.kcourb)));
		else
			return(chi2(z1,z2)/(1+z2));
	}
}

/****************************************************************/
/* Return the lens efficacity E=DA(LS) / DA(OS)
 *
 * If zl > zs, return 0. 
 * 
 * Global variables used :
 * - in distcosmo1() : C
 * - in distcosmo2() : C
 */
double dratio(double zl,double zs)
{
	if ( zl >= zs )
		return (0.);
	else
		return (distcosmo2(zl,zs)/distcosmo1(zs));
}

/**************************************************************
 * Return the luminosity distance DL(0,z) = (1+z)^2 * DA(0,z) 
 */
double dlumcosmo1(double z)
{
	return(distcosmo1(z)*(1.+z)*(1.+z));
}

/**************************************************************/

double distprime1(double z)
{
	extern  struct  g_cosmo C;
	double   g;
	
	if (C.omegaX==0.)
	{
		g=1./(1.+z);
		return(g*(g/gg(z)-distcosmo1(z)*2.*g));
	}
	else
	{
		return((distcosmo1(z+0.1)-distcosmo1(z))/0.1);
	}
}

/**************************************************************
 * Global variables used :
 * - in gg() : C
 * - in chi2() : C
 */
double distprime2(double z1,double z2)
{
	extern struct   g_cosmo C;
	double  g2; //g1
	
	if (C.omegaX==0.)
	{
		g2=1./(1.+z2);
		return(g2*(g2/gg(z2)-distcosmo2(z1,z2)*2.*g2));
	}
	else
	{
		return((distcosmo2(z1,z2+0.1)-distcosmo2(z1,z2))/0.1);
	}
}

/**************************************************************/

double dratioprime(double zl,double zs)
{
	double  ds;
	
	ds=distcosmo1(zs);
	return( (distprime2(zl,zs)-distcosmo2(zl,zs)*distprime1(zs)/ds)/ds );
}

/**************************************************************/
/* Global variables used :
 * - in chiz() : C
 * */
static double gg(double z)
{
	extern struct g_cosmo C;
	
	return(sqrt(1.+C.omegaM*z));
}
/**************************************************************
 * Global variables used :
 * - none
 */

static double sk(double x,double k)
{
	if (k>0)
		return(sin(x));
	else if (k<0)
		return(sinh(x));
	else
		return(x);
}

/**************************************************************
 * Return 1 / H(z) multiplied by H0
 *   chiz(z) / H0 = 1 / H(z) = 1/H0/(1+z)/sqrt( sum(i, Omega_i*(1+z)^(3*(w_i + 1) ) )
 *   with w_M = 0 and w_k = -1/3
 * 
 * Global variables used :
 * - C
 * */
double chiz(double z)
{
	extern struct g_cosmo C;
	double x;
	
	x=1+z;
	x = -x*x*C.kcourb+x*x*x*C.omegaM+pow(x,3*(1+C.wX))*C.omegaX;
	if( x <= 0 )
	{
		fprintf( stderr, "ERROR : H^2(z)<=0 produced (z,omegaM,omegaX,wX) = (%.3lf,%.3lf,%.3lf,%.3lf)\n",
			z, C.omegaM, C.omegaX, C.wX );
		exit(-1); 
	}
	return( 1./sqrt(x) );
}
/**************************************************************
 * Return the proper distance Dp(0,z) divided by c/H0
 *  chi1(z) * c/H0 = Dp(0,z) = integral( 0, z, c*dz / H(z) )
 * 
 * Global variables used :
 * - in chiz() : C
 * */
static double chi1(double z)
{
	return(qromb(chiz,0.,z));
}
/**************************************************************
 * Return the proper distance Dp(z1,z2) divided by c/H0
 *  chi2(z) * c/H0 = Dp(z1,z2) = integral( z1, z2, c*dz / H(z) )
 * 
 * Global variables used :
 * - in chiz() : C
 */
static double chi2(double z1,double z2)
{
	return(qromb(chiz,z1,z2));
}

