#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "constant.h"
#include "dimension.h"
#include "structure.h"
#include "fonction.h"

/********************************************************/
/*		fonction: recale			*/
/*		auteur: EJ				*/
/********************************************************
 * Rescale the cube values to their corresponding parameter values
 * in Lenstool. Each atom corresponds to a specific clump.
 * Fill the lens global variable.
 * 
 * Return 1 if there are the number of clumps in the cubes is equal to the 
 * number of lenses, 0 otherwise which means the object has to be rejected.
 * 
 * Parameters :
 * - cubes[clump][param] : the cubes containing the parameters for all clumps
 * - nclumps : the number of clumps in the cubes
 * - npar : the number of parameters
 *  
 * Global variables used :
 */
 
#define POTENTIAL 0
#define COSMO 1
#define ZMLIMIT 2
#define POT 3
#define SIGPOS 4

int rescaleCube_NAtoms(double **cubes, int nclumps, int npar)
{
 	extern struct g_grille	G;
 	extern struct g_mode	M;
 	
 	if( G.nlens != nclumps )
 	{
 		NPRINTF(stderr, "ERROR: Sample rejected (%d clumps)\n", nclumps);
 		return 0;
 	}
 	
	//TODO 
	return 1;
}
 
static double 	prior(int k, double val, double x1, double x2);

/* Rescale the lens parameter ipx for the lens ilens and test the
 * physical validy of the rescaled values.
 * 
 * Parameters :
 * -ilens : lens index in lens global variable
 * -ipx : parameter index in the Param enumerated type
 * -val : value from the cube in the [0..1] range
 * 
 * Return 1 if it's ok, 0 if it's not physicaly valid
 *  
 * Global variables used :
 * -lmin, lmax
 */
static int rescaleParam(int ilens, int type, int ipx, double val)
{
	extern struct pot	      lens[], lmin[], lmax[];
	extern struct pot       clmin, clmax;
	extern struct z_lim     zlim[];
	extern struct galaxie   multi[NIMAX][NIMAX];
	extern struct g_pot     P;
	extern struct g_image   I;

	extern int block[][NPAMAX];
	extern int cblock[NPAMAX];
	int i,j;
	int nimages;
	char limages[ZMBOUND][IDSIZE];
	
	int valid = 1;
	
	if( type == POTENTIAL || type == COSMO )
	{
		switch(ipx)
		{
			case(CX):
				val = prior( block[ilens][CX], val, 
					lmin[ilens].C.x, lmax[ilens].C.x );
				break;
			case(CY):
				val = prior( block[ilens][CY], val, 
					lmin[ilens].C.y, lmax[ilens].C.y );
				break;
			case(EPOT):
				val = prior( block[ilens][EPOT], val, 
					lmin[ilens].emass, lmax[ilens].emass );
				if( val < 0 ) valid = 0;				
				break;
			case(THETA):
				val = prior( block[ilens][THETA], val, 
					lmin[ilens].theta, lmax[ilens].theta );
				break;
			case(RC):
				val = prior( block[ilens][RC], val, 
					lmin[ilens].rc, lmax[ilens].rc );
				if( val < 0 ) valid = 0;
				break;
			case(B0):
				// rescaling to sigma instead of B0 to
				// keep the shape of the input prior.
				val = prior( block[ilens][B0], val, 
					lmin[ilens].sigma, lmax[ilens].sigma );
				if( val < 0 ) 
					valid = 0; 
				else
					lens[ilens].sigma = val;	
				break;
			case(ALPHA):
				val = prior( block[ilens][ALPHA], val, 
					lmin[ilens].alpha, lmax[ilens].alpha );
				break;
			case(BETA):
				val = prior( block[ilens][BETA], val, 
					lmin[ilens].beta, lmax[ilens].beta );
				break;																																										
			case(RCUT):
				val = prior( block[ilens][RCUT], val, 
					lmin[ilens].rcut, lmax[ilens].rcut );
				if( val < 0 ) valid = 0;					
				break;
			case(PMASS):
				val = prior( block[ilens][PMASS], val,
					lmin[ilens].pmass, lmax[ilens].pmass );
				if( val < 0 ) valid = 0;
				break;				
			case(MASSE):
				val = prior( block[ilens][MASSE], val,
					lmin[ilens].masse, lmax[ilens].masse );
				if( val < 0 ) valid = 0;
				break;
			case(OMEGAM):
				val = prior( cblock[OMEGAM], val, 
					clmin.omegaM, clmax.omegaM );
				// Test physical validity. beause: H(z)^2 can become negative 
				// in the corner OmegaX > 1 and OmegaM ~ 0, ie Universes with "rebond"
				if( val < 0 ) valid = 0;
				break;
			case(OMEGAX):
				val = prior( cblock[OMEGAX], val, 
					clmin.omegaX, clmax.omegaX );
				if( val > 1 ) valid = 0;
				break;
			case(WX):
				val = prior( cblock[WX], val, 
					clmin.wX, clmax.wX );
				break;
		};
		if( valid && ipx != B0 )
			o_set_lens( ilens, ipx, val );
	}
	

	// zm_limit section
	if( type == ZMLIMIT )
	{
		val = prior( zlim[ipx].bk, val, zlim[ipx].min, zlim[ipx].max );
		if( val < 0 ) 
			valid = 0;
		else
		{
			nimages = splitzmlimit( zlim[ipx].n, limages );
			// for all the families in zlim[ipx].n
			for( j = 0; j < nimages; j++ )
			{
				// look for this family in the multi array
				i = 0;
				while( indexCmp( multi[i][0].n, limages[j] ) ) i++;
				multi[i][0].z = val;
			}
		} 
	}
	
	// rescale the potfile parameters rcut and sigma
	if( type == POT )
	{
		switch(ipx)
		{
			case(0):
				P.cut = prior( P.ircut, val, P.cut1, P.cut2 );
				if( P.cut <= 0 ) valid = 0;
				break;
			case(1):
				P.sigma = prior( P.isigma, val, P.sigma1, P.sigma2 );
				if( P.sigma <=0 ) valid = 0;
				break;			
		};		
	}
	
	// rescale the sigposAs noise
	if( type == SIGPOS )
	{
		switch(ipx)
		{
			case(0): // sigposArsec
				val = prior( 3, val, I.sigposAs,I.dsigposAs );
				if( val <= 0 ) valid = 0;
				I.sig2pos = val*val;
				break;
			case(1): // sigell
				val = prior( 3, val, I.sigell,I.dsigell );
				if( val <= 0 ) valid = 0;
				I.sig2ell = val*val;
				break;			
		}
	}
	
	return valid;
}

static double uniformPrior(double val, double min, double max);
static double gaussianPrior(double val, double mu, double sigma);

/* Prior distribution functions : val is a deviate from the unit Cube.
 * 
 * Return the corresponding deviate drawn from the desired distribution 
 * (labelled k).
 */
static double prior(int k, double val, double x1, double x2)
{
	switch(k)
	{
		case(1) :
			return uniformPrior(val, x1, x2);
			break;
		case(3) :
			return gaussianPrior(val, x1, x2);
			break;
		default :
			fprintf( stderr, "ERROR : prior index %d undefined.\n", k );
			exit(-1);
	}
	
	return 0; //to avoid the compiler warning.
}

/* Uniform [0:1] -> Uniform[x1:x2]
 */
static double uniformPrior(double val, double min, double max)
{
	return 	min + val * (max - min);
}

static double	dierfc(double y);

/* Uniform [0:1] -> Gaussian[mean=mu, variance=sigma**2]
 */
static double gaussianPrior(double val, double mu, double sigma)
{
	double	sqrtTwo = 1.414213562;

	return mu + sigma*sqrtTwo*dierfc(2. * (1. - val) );
}

/* Inverse of error function in double precision
 */
static double dierfc(double y)
{
	double
		qa = 9.16461398268964e-01,
		qb = 2.31729200323405e-01,
		qc = 4.88826640273108e-01,
		qd = 1.24610454613712e-01,
		q0 = 4.99999303439796e-01,
		q1 = 1.16065025341614e-01,
		q2 = 1.50689047360223e-01,
		q3 = 2.69999308670029e-01,
		q4 = -7.28846765585675e-02,
		
		pa = 3.97886080735226000e+00,
		pb = 1.20782237635245222e-01,
		p0 = 2.44044510593190935e-01,
		p1 = 4.34397492331430115e-01,
		p2 = 6.86265948274097816e-01,
		p3 = 9.56464974744799006e-01,
		p4 = 1.16374581931560831e+00,
		p5 = 1.21448730779995237e+00,
		p6 = 1.05375024970847138e+00,
		p7 = 7.13657635868730364e-01,
		p8 = 3.16847638520135944e-01,
		p9 = 1.47297938331485121e-02,
		p10 = -1.05872177941595488e-01,
		p11 = -7.43424357241784861e-02,
		 
		p12 = 2.20995927012179067e-03,
		p13 = 3.46494207789099922e-02,
		p14 = 1.42961988697898018e-02,
		p15 = -1.18598117047771100e-02,
		p16 = -1.12749169332504870e-02,
		p17 = 3.39721910367775861e-03,
		p18 = 6.85649426074558612e-03,
		p19 = -7.71708358954120939e-04,
		p20 = -3.51287146129100025e-03,
		p21 = 1.05739299623423047e-04,
		p22 = 1.12648096188977922e-03;

	double infinity = 5.0;
	double z, w, u, s, t, x;
         
	if( y < 1.e-40 )
        return infinity;

	z = y;
	if( y > 1. ) z = 2 - y;
	w = qa - log10(z);
	u = sqrt(w);
	s = (qc + log10(u)) / w;
	t = 1. / (u + qb);
	x = u * (1. - s * (0.5 + s * qd)) -
		((((q4 * t + q3) * t + q2) * t + q1) * t + q0) * t;
	t = pa / (pa + x);
	u = t - 0.5;
	s = (((((((((p22 * u + p21) * u + p20) * u +
		p19) * u + p18) * u + p17) * u + p16) * u +
		p15) * u + p14) * u + p13) * u + p12;
	s = ((((((((((((s * u + p11) * u + p10) * u +
		p9) * u + p8) * u + p7) * u + p6) * u + p5) * u +
		p4) * u + p3) * u + p2) * u + p1) * u + p0) * t -
		z * exp(x * x - pb);
	x = x + s * (1. + x * s);
	x = x + s * (1. + x * s);
	if( y > 1.) x = -x;

    return x;
}

 /********************************************************
 * Rescale the cube values to their corresponding parameter values
 * in Lenstool. A single atom contains all the clumps
 * Fill the lens global variable. 
 * 
 * Return 1 if the rescaled parameters have significant meaning, 
 * 0 otherwise.
 * 
 * Parameters :
 * - cube[param] : the cubes containing the parameters for all clumps
 * - npar : the number of parameters
 *  
 * Global variables used :
 * - 
 */
int rescaleCube_1Atom(double *cube, int npar)
{
 	extern struct g_grille	G;
// 	extern struct g_source	S;
	extern struct g_cosmo	C;
 	extern struct g_mode	M;
	extern struct g_image	I;
	extern struct g_pot	P;
	extern struct galaxie	arclet[NAMAX];
	extern struct pot		lens[];
	extern struct galaxie	multi[NIMAX][NIMAX];
	extern struct z_lim	zlim[];
	extern struct cline     cl[];
	
	extern int narclet;	
 	extern int block[][NPAMAX];
 	extern int cblock[NPAMAX];
 	
 	int ipar; 	/*index of the current parameter in the atom*/
 	int ipx;	/*index of the current parameter in the ipot structure*/
 	int ilens;	/*index of the clump in the lens[] global variable*/
 	int valid; 		// Accept or not this sample
 	int computedr;	// flag
 	int potparam;	// flag
 	int i;
// 	double z;
 	
	valid = 1;		// OK 
 	computedr = 0;	// do not recompute dlsds ratio for multiple images
 	potparam = 0;	// potfile parameters are not modified
 	
 	// Rescale de potentials
 	ipar=0;
 	for( ilens = 0; ilens < G.no_lens; ilens++ )
 	{
 		// check over the lens parameters those which are to be optimised
		for( ipx = CX; ipx <= PMASS; ipx++ )
			if( block[ilens][ipx] != 0 )
				valid *= rescaleParam(ilens, POTENTIAL, ipx,(double)cube[ipar++]);

		// rescale B0
		set_dynamics(ilens);
		
		if( lens[ilens].rc > lens[ilens].rcut )
			valid = 0;
	}
	
	// rescale the cosmological parameters
	for( ipx = OMEGAM; ipx <= WX; ipx++ )
		if( cblock[ipx] != 0 )
		{
			valid *= rescaleParam(0, COSMO, ipx,(double)cube[ipar++]);
			computedr = 1;
		}
	
	// rescale the z_m_limit redshifts
	for( i = 0; i < I.nzlim; i++ )
		if( zlim[i].opt != 0 )
		{
			valid *= rescaleParam(0, ZMLIMIT, i,(double)cube[ipar++]);
			computedr = 1;
		}	

	// rescale the pot parameters
	if( P.ircut != 0 )
	{
		valid *= rescaleParam(0, POT, 0, (double)cube[ipar++]);
		potparam = 1;
	}
	if( P.isigma != 0 )
	{
		valid *= rescaleParam(0, POT, 1, (double)cube[ipar++]);
		potparam = 1;
	}
	
	// rescale the noise
	if( I.dsigposAs != -1. )
		valid *= rescaleParam(0, SIGPOS, 0, (double)cube[ipar++]);
	if( I.dsigell != -1. )
		valid *= rescaleParam(0, SIGPOS, 1, (double)cube[ipar++]);

	// recompute the dlsds ratios if necessary
	if( computedr  && valid ) 
	{
		if( C.kcourb == 0. )
			C.omegaX = 1 - C.omegaM;
		else
			C.kcourb = C.omegaM + C.omegaX - 1;

//		z = multi[I.n_mult-1][0].z + 1;
		// so that the highest redshift can exist in such a univers
//		if( -z*z*C.kcourb+z*z*z*C.omegaM+C.omegaX*pow(z, 3*(1+C.wX)) > 0 )

		for( i = 0 ; i < I.n_mult; i++ )
			multi[i][0].dr=dratio(lens[0].z,multi[i][0].z);
			
		//...the same for the arclets
		for( i = 0 ; i < narclet; i++ )
			arclet[i].dr = dratio(lens[0].z,arclet[i].z);
			
		// ...same for the critical line dlsds
		for( i = 0 ; i < I.npcl; i++ )
			cl[i].dlsds = dratio(lens[0].z,cl[i].z);
	}
	
	// recompute the potfile if necessary
	if( potparam && valid )
	{
		// scale the potfile clumps
		o_scale_pot();
		
		// Reset the grad2 variable of the multiple images to 0
		// so that it is computed again for all the clumps 
		// in e_grad_gal() and e_grad2_gal()
		/* Already done in o_scale_pot()
		for( i = 0 ; i < I.n_mult ; i++ )
			for( j = 0 ; j < I.mult[i]; j++ )
			{
				multi[i][j].grad2.a = multi[i][j].grad2.c = 0;
				multi[i][j].grad.x = multi[i][j].grad.y = 0;
			}
			
		for( i = 0 ; i< narclet ; i++ )
		{
			arclet[i].grad2.a = arclet[i].grad2.c = 0;
			arclet[i].grad.x = arclet[i].grad.y = 0;
		}
		*/	
	}
	
	return valid;
}
 
